import { jsx as _jsx } from "react/jsx-runtime";
import { memo, useCallback, useEffect, useRef } from "react";
import shallow from "zustand/shallow";
import { usePlotStore } from ".";
import { extend } from "@react-three/fiber";
import { MeshLine, MeshLineMaterial } from 'meshline';
extend({ MeshLine, MeshLineMaterial });
import { makeCurve, updateCurve } from "./plots/curve";
import { makeImplicit, updateImplicit } from "./plots/implicit";
import { makeOneForm, updateOneForm } from "./plots/oneForm";
import { makeParametric, updateParametric } from "./plots/parametric";
import { Point } from "./plots/Point";
import { makeVector, updateVector } from "./plots/vector";
import { makeVectorField, updateVectorField } from "./plots/vectorField";
const TWOPI = 2 * Math.PI;
const vars = ["x", "y", "z"];
export const ExpressionPlot = memo(function ExpressionPlot(props) {
    var _a, _b, _c, _d, _e;
    const useStore = usePlotStore();
    const expr = useRef();
    expr.current = useStore.getState().expressions.find(_ => _.id === props.id);
    const ref = useRef();
    switch (true) {
        case !!expr.current.point:
            return _jsx(Point, {}, void 0);
    }
    useEffect(() => {
        const desubs = [];
        desubs.push(useStore.subscribe((hidden) => {
            if (!ref.current)
                return;
            ref.current.visible = !(hidden !== null && hidden !== void 0 ? hidden : false);
        }, state => state.expressions.find(_ => _.id === props.id).hidden));
        desubs.push(useStore.subscribe((opacity) => {
            if (!ref.current)
                return;
            if (opacity !== 1) {
                ref.current.material.opacity = opacity;
                ref.current.material.transparent = true;
            }
            else {
                ref.current.material.transparent = false;
            }
        }, state => state.expressions.find(_ => _.id === props.id).opacity));
        desubs.push(useStore.subscribe((e) => {
            expr.current = e;
            if (!ref.current)
                return;
            updateGeometry(e);
            if (!e.latex)
                return;
            let $_;
            if ($_ = e.latex.replace(/\s+/g, "").match(/^([xyz])=(-?[\d.]+)$/)) {
                const v = $_[1];
                const val = parseFloat($_[2]);
                ref.current.position["set" + v.toUpperCase()](val);
                if (v === "x") {
                    ref.current.rotation.set(0, TWOPI / 4, 0);
                }
                else if (v === "y") {
                    ref.current.rotation.set(TWOPI / 4, 0, 0);
                }
                else if (v === "z") {
                    ref.current.rotation.set(0, 0, 0);
                }
            }
        }, state => state.expressions.find(_ => _.id === props.id), shallow));
        if (expr.current.vars) {
            desubs.push(useStore.subscribe(() => {
                updateGeometry(expr.current);
            }, state => e.vars.map(v => state.vars[v]), shallow));
        }
        return () => desubs.map(_ => _());
    }, []);
    const updateGeometry = useCallback((e) => {
        if (e.curve) {
            updateCurve(e, useStore.getState(), ref);
        }
        else if (e.parametric) {
            updateParametric(e, ref);
        }
        else if (e.implicit) {
            updateImplicit(e, useStore.getState(), ref);
        }
        else if (e.vector) {
            updateVector({ e, useStore, ref });
        }
        else if (e.vectorField) {
            updateVectorField({ e, ref, useStore });
        }
        else if (e.oneForm) {
            updateOneForm({ e, ref, useStore });
        }
    }, []);
    const e = expr.current;
    if (!e) {
        return null;
    }
    let geometry, material, $_;
    const attrs = {
        name: props.id,
        visible: !((_a = e.hidden) !== null && _a !== void 0 ? _a : false),
        opacity: (_b = e.opacity) !== null && _b !== void 0 ? _b : 1,
        transparent: e.opacity !== 1,
        position: [0, 0, 0],
        rotation: [0, 0, 0]
    };
    const materialAttrs = {
        color: e.color ? new THREE.Color(e.color).convertSRGBToLinear() : 0xFFFFFF,
        opacity: (_c = e.opacity) !== null && _c !== void 0 ? _c : 1,
        side: THREE.DoubleSide,
        transparent: e.opacity !== 1,
    };
    (_d = e.material) !== null && _d !== void 0 ? _d : (e.material = "toon");
    const materialConstructor = THREE["Mesh" + e.material[0].toUpperCase() + e.material.slice(1) + "Material"];
    if (!materialConstructor) {
        console.error("Invalid material:", e.material);
    }
    else {
        material = new materialConstructor(materialAttrs);
    }
    if (e.curve) {
        geometry = makeCurve(e, useStore.getState());
    }
    else if (e.implicit) {
        geometry = makeImplicit(e, useStore.getState());
    }
    else if (e.oneForm) {
        return makeOneForm({ e, attrs, ref, useStore });
    }
    else if (e.parametric) {
        geometry = makeParametric(e);
    }
    else if (e.point) {
        attrs.position = e.point;
        geometry = new THREE.SphereGeometry((_e = e.pointSize) !== null && _e !== void 0 ? _e : 0.01, 32, 32);
    }
    else if (e.vector) {
        return makeVector({ e, attrs, ref, useStore });
    }
    else if (e.vectorField) {
        return makeVectorField({ e, attrs, ref, useStore });
    }
    else if ($_ = e.latex.replace(/\s+/g, "").match(/([xyz])=(-?[\d.+])/)) {
        const v = $_[1];
        const val = parseFloat($_[2]);
        geometry = new THREE.PlaneGeometry(10, 10);
        if (v === "x") {
            attrs.rotation = [0, TWOPI / 4, 0];
        }
        else if (v === "y") {
            attrs.rotation = [TWOPI / 4, 0, 0];
        }
        attrs.position[vars.indexOf(v)] = val;
    }
    else {
        return null;
    }
    return (_jsx("mesh", Object.assign({}, attrs, { geometry: geometry, material: material, ref: ref }), void 0));
}, () => true);
